#include <config.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <gio/gio.h>

#include "hrn.h"
#include "hrn-queue-item-factory.h"
#include "hrn-queue-item.h"
#include "hrn-source-manager.h"
#include "hrn-queue.h"
#include "hrn-queue-header.h"

enum {
    PROP_0,
    PROP_QUEUE,
    PROP_MANAGER
};

enum {
    LAST_SIGNAL,
};

struct _HrnQueuePrivate {
    ClutterModel *model;
    HrnQueueItemFactory *factory;
    HrnQueueHeader *header;
    HrnSourceManager *manager;
    BrQueue *queue;

    NbtkWidget *scroll;
    NbtkWidget *listview;
};

#define QUEUE_WIDTH 190
#define QUEUE_HEIGHT 200

enum {
    MODEL_ITEM,
    MODEL_URI,
};

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), HRN_TYPE_QUEUE, HrnQueuePrivate))
G_DEFINE_TYPE (HrnQueue, hrn_queue, NBTK_TYPE_BIN);
static guint32 signals[LAST_SIGNAL] = {0,};

static void
hrn_queue_finalize (GObject *object)
{
    G_OBJECT_CLASS (hrn_queue_parent_class)->finalize (object);
}

static void
hrn_queue_dispose (GObject *object)
{
    G_OBJECT_CLASS (hrn_queue_parent_class)->dispose (object);
}

static void
clear_header (HrnQueue *queue)
{
    HrnQueuePrivate *priv = queue->priv;

    g_object_set (priv->header,
                  "primary", _("Music Playqueue"),
                  "secondary", _("Add music to start"),
                  "thumbnail", NULL,
                  NULL);
}

static void
brq_uri_added (BrQueue    *brq,
               const char *uri,
               int         position,
               HrnQueue   *queue)
{
    HrnQueuePrivate *priv = queue->priv;
    BklItem *item;

    item = hrn_source_manager_get_item_for_uri (priv->manager, uri, NULL);
    clutter_model_insert (priv->model, position,
                          MODEL_ITEM, item,
                          MODEL_URI, uri,
                          -1);
    hrn_queue_header_expand (priv->header, TRUE);
}

static void
brq_uri_removed (BrQueue    *brq,
                 const char *uri,
                 int         position,
                 HrnQueue   *queue)
{
    HrnQueuePrivate *priv = queue->priv;

    clutter_model_remove (priv->model, position);

    if (clutter_model_get_n_rows (priv->model) == 0) {
        clear_header (queue);
    }
}

static void
brq_position_changed (BrQueue  *brq,
                      double    position,
                      HrnQueue *queue)
{
    HrnQueuePrivate *priv = queue->priv;

    hrn_queue_header_set_progress (priv->header, position);
}

static void
set_now_playing (HrnQueue   *queue,
                 BklItem    *item,
                 const char *uri)
{
    HrnQueuePrivate *priv = queue->priv;

    if (item == NULL && uri) {
        char *filename, *unesc;

        filename = g_path_get_basename (uri);
        unesc = g_uri_unescape_string (filename, NULL);
        g_free (filename);

        g_object_set (priv->header,
                      "primary", unesc,
                      "secondary", "",
                      NULL);
        g_free (unesc);
    } else if (item) {
        const char *primary, *secondary;
        GPtrArray *artists;

        primary = bkl_item_audio_get_title ((BklItemAudio *) item);
        artists = bkl_item_audio_get_artists ((BklItemAudio *) item);
        if (artists) {
            secondary = artists->pdata[0];
        } else {
            secondary = _("Unknown Artist");
        }

        g_object_set (priv->header,
                      "primary", primary ? primary : _("Unknown"),
                      "secondary", secondary,
                      NULL);
    } else {
        /* Erk? */
        g_object_set (priv->header,
                      "primary", _("Unknown"),
                      "secondary", _("Unknown Artist"),
                      NULL);
    }
}

static void
brq_index_changed (BrQueue  *brq,
                   int       idx,
                   HrnQueue *queue)
{
    HrnQueuePrivate *priv = queue->priv;
    ClutterModelIter *iter;
    BklItem *item;
    char *uri;

    iter = clutter_model_get_iter_at_row (priv->model, idx);
    if (iter == NULL) {
        g_warning ("Error getting iter at row %d", idx);
        return;
    }

    clutter_model_iter_get (iter,
                            MODEL_ITEM, &item,
                            MODEL_URI, &uri,
                            -1);

    set_now_playing (queue, item, uri);
    if (item) {
        g_object_unref (item);
    }
    g_free (uri);
}

static void
get_name_reply (BrQueue      *brq,
                const char   *name,
                const GError *error,
                gpointer      data)
{
}

static void
list_uris_reply (BrQueue       *brq,
                 const char   **uris,
                 const GError  *error,
                 gpointer       data)
{
    HrnQueue *queue = (HrnQueue *) data;
    HrnQueuePrivate *priv = queue->priv;
    int i;

    if (error != NULL) {
        g_warning ("Error listing queue uris: %s", error->message);
        return;
    }

    for (i = 0; uris[i]; i++) {
        BklItem *item = hrn_source_manager_get_item_for_uri (priv->manager,
                                                             uris[i], NULL);
        clutter_model_append (priv->model,
                              MODEL_ITEM, item,
                              MODEL_URI, uris[i],
                              -1);
    }
    hrn_queue_header_expand (priv->header, TRUE);
}

static void
get_index_reply (BrQueue      *brq,
                 int           idx,
                 const GError *error,
                 gpointer      data)
{
    if (error != NULL) {
        g_warning ("Error getting index: %s", error->message);
        return;
    }

    brq_index_changed (brq, idx, data);
}

static void
source_ready_cb (HrnSourceManager *manager,
                 HrnQueue         *queue)
{
    HrnQueuePrivate *priv = queue->priv;

    if (priv->queue) {
        br_queue_list_uris (priv->queue, list_uris_reply, queue);
        br_queue_get_index (priv->queue, get_index_reply, queue);
    }
}

static void
hrn_queue_set_property (GObject      *object,
                        guint         prop_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
    HrnQueue *self = (HrnQueue *) object;
    HrnQueuePrivate *priv = self->priv;

    switch (prop_id) {
    case PROP_QUEUE:
        priv->queue = g_value_dup_object (value);
        g_signal_connect (priv->queue, "uri-added",
                          G_CALLBACK (brq_uri_added), self);
        g_signal_connect (priv->queue, "uri-removed",
                          G_CALLBACK (brq_uri_removed), self);
        g_signal_connect (priv->queue, "index-changed",
                          G_CALLBACK (brq_index_changed), self);
        g_signal_connect (priv->queue, "position-changed",
                          G_CALLBACK (brq_position_changed), self);

        /* br_queue_get_name (priv->queue, get_name_reply, self); */
        break;

    case PROP_MANAGER:
        priv->manager = g_value_dup_object (value);

        if (hrn_source_manager_is_ready (priv->manager)) {
            source_ready_cb (priv->manager, self);
        } else {
            g_signal_connect (priv->manager, "ready",
                              G_CALLBACK (source_ready_cb), self);
        }
        break;

    default:
        break;
    }
}

static void
hrn_queue_get_property (GObject    *object,
                        guint       prop_id,
                        GValue     *value,
                        GParamSpec *pspec)
{
    switch (prop_id) {

    default:
        break;
    }
}

#if 0
static GObject *
hrn_queue_constructor (GType                  type,
                       guint                  n_params,
                       GObjectConstructParam *params)
{
    HrnQueuePrivate *priv;
    GObject *object;
    HrnQueue *queue;

    object = G_OBJECT_CLASS (hrn_queue_parent_class)->constructor (type,
                                                                   n_params,
                                                                   params);
    queue = (HrnQueue *) object;
    priv = queue->priv;

    return object;
}
#endif

static void
hrn_queue_class_init (HrnQueueClass *klass)
{
    GObjectClass *o_class = (GObjectClass *) klass;

    o_class->dispose = hrn_queue_dispose;
    o_class->finalize = hrn_queue_finalize;
    o_class->set_property = hrn_queue_set_property;
    o_class->get_property = hrn_queue_get_property;
    /* o_class->constructor = hrn_queue_constructor; */

    g_type_class_add_private (klass, sizeof (HrnQueuePrivate));

    g_object_class_install_property (o_class, PROP_QUEUE,
                                     g_param_spec_object ("queue", "", "",
                                                          BR_TYPE_QUEUE,
                                                          G_PARAM_STATIC_STRINGS |
                                                          G_PARAM_CONSTRUCT_ONLY |
                                                          G_PARAM_WRITABLE));
    g_object_class_install_property (o_class, PROP_MANAGER,
                                     g_param_spec_object ("manager", "", "",
                                                          HRN_TYPE_SOURCE_MANAGER,
                                                          G_PARAM_STATIC_STRINGS |
                                                          G_PARAM_CONSTRUCT_ONLY |
                                                          G_PARAM_WRITABLE));
}


static void
progress_notify_cb (HrnQueueHeader *header,
                    double          position,
                    HrnQueue       *queue)
{
    HrnQueuePrivate *priv = queue->priv;

    br_queue_set_position (priv->queue, position);
}

static void
expanded_clicked_cb (HrnQueueHeader *header,
                     gboolean        expanded,
                     HrnQueue       *queue)
{
    HrnQueuePrivate *priv = queue->priv;

    if (expanded) {
        clutter_actor_animate ((ClutterActor *) queue,
                               CLUTTER_EASE_IN_OUT_CUBIC, 350,
                               "y", -200.0,
                               "height", 279.0,
                               NULL);
        clutter_actor_show ((ClutterActor *) priv->scroll);
    } else {
        clutter_actor_animate ((ClutterActor *) queue,
                               CLUTTER_EASE_IN_OUT_CUBIC, 350,
                               "y", 0.0,
                               "height", 72.0,
                               "signal-swapped::completed",
                               clutter_actor_hide, priv->scroll,
                               NULL);
    }
}

static void
item_clicked_cb (HrnQueueItemFactory *factory,
                 HrnQueueItem        *item,
                 HrnQueue            *queue)
{
    HrnQueuePrivate *priv = queue->priv;
    char *uri;
    ClutterModelIter *iter;
    int row = -1;

    g_object_get (item,
                  "uri", &uri,
                  NULL);

    if (uri == NULL) {
        g_warning ("No uri for item");
        return;
    }

    /* Look through the model to find the row
       FIXME: there really should have been a better way to do this */
    iter = clutter_model_get_first_iter (priv->model);
    while (!clutter_model_iter_is_last (iter)) {
        char *model_uri;

        clutter_model_iter_get (iter,
                                MODEL_URI, &model_uri,
                                -1);

        if (model_uri == NULL) {
            g_warning ("No uri for model");
            iter = clutter_model_iter_next (iter);
            continue;
        }

        if (g_str_equal (model_uri, uri)) {
            row = clutter_model_iter_get_row (iter);
            g_free (model_uri);
            break;
        }

        iter = clutter_model_iter_next (iter);
    }

    g_free (uri);

    if (row == -1) {
        return;
    }

    br_queue_set_index (priv->queue, row);
}

static void
hrn_queue_init (HrnQueue *self)
{
    HrnQueuePrivate *priv = GET_PRIVATE (self);
    ClutterActor *layout;

    self->priv = priv;
    priv->model = clutter_list_model_new (2, BKL_TYPE_ITEM, "item",
                                          G_TYPE_STRING, "uri");

    layout = (ClutterActor *) nbtk_box_layout_new ();
    nbtk_box_layout_set_vertical ((NbtkBoxLayout *) layout, TRUE);
    clutter_container_add_actor ((ClutterContainer *) self, layout);

    priv->header = g_object_new (HRN_TYPE_QUEUE_HEADER, NULL);
    g_signal_connect (priv->header, "expand",
                      G_CALLBACK (expanded_clicked_cb), self);
    g_signal_connect (priv->header, "progress",
                      G_CALLBACK (progress_notify_cb), self);
    clutter_container_add_actor ((ClutterContainer *) layout,
                                 (ClutterActor *) priv->header);
    clear_header (self);

    priv->scroll = (NbtkWidget *) nbtk_scroll_view_new ();
    clutter_container_add_actor ((ClutterContainer *) layout,
                                 (ClutterActor *) priv->scroll);
    clutter_actor_set_size ((ClutterActor *) priv->scroll,
                            QUEUE_WIDTH, QUEUE_HEIGHT);
    priv->listview = nbtk_list_view_new ();
    clutter_container_add_actor ((ClutterContainer *) priv->scroll,
                                 (ClutterActor *) priv->listview);
    nbtk_list_view_set_model (NBTK_LIST_VIEW (priv->listview), priv->model);

    priv->factory = hrn_queue_item_factory_new ();
    g_signal_connect (priv->factory, "item-clicked",
                      G_CALLBACK (item_clicked_cb), self);
    nbtk_list_view_set_factory (NBTK_LIST_VIEW (priv->listview),
                                NBTK_ITEM_FACTORY (priv->factory));

    nbtk_list_view_add_attribute (NBTK_LIST_VIEW (priv->listview), "item", 0);
    nbtk_list_view_add_attribute (NBTK_LIST_VIEW (priv->listview), "uri", 1);
}

HrnQueue *
hrn_queue_new (BrQueue          *queue,
               HrnSourceManager *manager)
{
    HrnQueue *q;

    q = g_object_new (HRN_TYPE_QUEUE,
                      "queue", queue,
                      "manager", manager,
                      NULL);

    return q;
}


void
hrn_queue_play_uri_now (HrnQueue    *queue,
                        const gchar *uri)
{
  HrnQueuePrivate *priv = queue->priv;
  const char *uris[2] = {0, }, *mimetypes[2] = {0, };
  GFile *file;
  char *mimetype;

  file = g_file_new_for_commandline_arg (uri);

  mimetype = hrn_resolve_mimetype (uri);
  uris[0] = g_file_get_uri (file);
  mimetypes[0] = mimetype;

  br_queue_append_uris (priv->queue, 1, uris, mimetypes);
  br_queue_play (priv->queue);

  g_free (mimetype);
  g_object_unref (file);
}

void
hrn_queue_play_now (HrnQueue *queue,
                    BklItem  *item)
{
  HrnQueuePrivate *priv = queue->priv;
  const char *uris[2] = {0, }, *mimetypes[2] = {0, };

  uris[0] = bkl_item_get_uri (item);
  mimetypes[0] = bkl_item_get_mimetype (item);

  br_queue_append_uris (priv->queue, 1, uris, mimetypes);
  br_queue_set_index (priv->queue, BR_QUEUE_INDEX_END);
  br_queue_play (priv->queue);
}
